/**********************************/
/*                                */
/*     db.h                       */
/*                                */
/*     database and schema        */
/*     definitions                */
/*                                */
/*     Alan Dix                   */
/*                                */
/*     26/11/1993                 */
/*                                */
/**********************************/

#ifndef DB_H
#define DB_H

#include <stdio.h>

typedef enum { ET_str, ET_num, ET_none } elem_type;

typedef struct schema_elem_st {
        char * name;
        elem_type type;
        int  size;
    } schema_elem, attribute;

typedef struct schema_st {
        int  size;
        attribute * elems;
    } schema;

typedef struct class_st {
        attribute det;
        int nos_classes;
        char **names;
    } class_sch;

typedef union {
        char * str;
        int    num;
    } tuple_elem;

typedef tuple_elem *tuple;

typedef struct tup_st {
        struct tup_st *next;
        tuple elem;
    } tuple_list;

typedef struct table_st {
        schema *sch;
        int  size;
        tuple_list *tups;
        tuple_list **last;
    } table;

typedef struct curr_st {
        table *tb;
        int    pos;
        tuple_list *tup;
    } currency;

typedef int attr_no;
typedef int tuple_no;
typedef int class_no;

#define NUM_MAX 0x7FFF     /* 16 bit ints */
/* #define NUM_MAX 0x7FFFFFFF  */   /* 16 bit ints */
#define NUM_MIN -NUM_MAX

#define null_num NUM_MIN-1
extern char *null_str;


void new_schema( schema *sch, int size );
schema *mk_new_schema( int size );
schema *copy_schema( schema *old_sch );
int equal_schema( schema *s1, schema *s2 );
void add_attribute_to_schema( schema *sch, attribute attrib  );
                                     /* updates existing schema */
                                     /* assumes malloced schema */
void add_schema_to_schema( schema *sch, schema *new  );
                                     /* updates existing schema */
                                     /* assumes malloced schema */
void set_attribute_in_schema( schema *sch, attr_no attr,
                              char * name, elem_type type, int  size );

tuple mk_tuple( int size );

void empty_table( table *tb, schema *sch );
          /* consumes reference to schema, copy if necessary */
table *mk_empty_table( schema *sch );
          /* consumes reference to schema, copy if necessary */
void free_table( table *tb );
      /* does NOT free individual tuples */
      /* does NOT free schema            */
void free_tuples_and_table( table *tb );
      /* does NOT free schema            */
void add_to_table( table *tb, tuple tup );
     /* assumes tuple is appropriate for schema */
void add_cur_to_table( table *tb, currency *cur );
     /* assumes currency is appropriate for schema */
void add_to_front_table( table *tb, tuple tup );
     /* assumes tuple is appropriate for schema */
void add_cur_to_front_table( table *tb, currency *cur );
     /* assumes currency is appropriate for schema */
void append_table( table *front, table *back );
     /* updates front             */
     /* assumes tables compatable */
void prepend_table( table *front, table *back );
     /* updates back              */
     /* assumes tables compatable */
table *concat_table( table *front, table *back );
     /* updates neither           */
     /* assumes tables compatable */
table *add_attrib_to_table( table *tb, attribute attrib, tuple_elem init );
               /* mallocs completely new table  */
table *add_schema_to_table( table *tb, schema *new, tuple init );
               /* mallocs completely new table  */

int first_currency( currency *cr, table *tb );
int is_last_currency( currency *cr );
int next_currency( currency *cr );
int copy_currency( currency *to, currency *from );
tuple tuple_at_currency( currency *cr );
int  set_num_in_currency ( currency *cr, attr_no attr, int num );
int  set_str_in_currency ( currency *cr, attr_no attr, char *str );
int  get_num_in_currency ( currency *cr, attr_no attr );
char *get_str_in_currency ( currency *cr, attr_no attr );

int  is_attr_num( attr_no attr, schema *sch );
int  is_attr_str( attr_no attr, schema *sch );
attr_no  get_attr_no( char *name, schema *sch );
class_no get_class_no( char *name, class_sch *cl );
char *name_of_class( class_no cn, class_sch *cl );
char *name_of_attrib( attr_no attr, schema *sch );

int read_schema( FILE *fp, schema *sch );
int read_table( FILE *fp, table *tab, schema *sch );
int read_into_table( FILE *fp, table *tab );
void print_db( FILE *fp, table *tab );
       /*  nicely formatted with headings  */
void print_table( FILE *fp, table *tab );
       /*  no headings, for computer processing  */
class_sch *read_class_sch( FILE *fp );

#endif /*  DB_H  */





