/**********************************/
/*                                */
/*     db_io.h                    */
/*                                */
/*     database and schema        */
/*     I/O                        */
/*                                */
/*     Alan Dix                   */
/*                                */
/*     26/11/1993                 */
/*                                */
/**********************************/

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "mem.h"

#include "db.h"
#include "gen_io.h"

static int read_attribute_details( FILE *fp, attribute *atd );
static int read_schema_elem( FILE *fp, schema *sch, attr_no attr );
static int read_tuple( FILE *fp, tuple *tup, schema *sch );

int read_schema( FILE *fp, schema *sch )
{
    int i, ret;
    token my_tok = read_token(fp);
    if ( my_tok != TOK_NUM ) {
        read_error = "expecting number of attributes";
        return 0;
        }
    new_schema( sch, tok_num );
/*     sch->size = tok_num;    */
/*     sch->elems = (schema_elem*)mem_malloc( sch->size * sizeof(schema_elem) );  */
    for ( i=0; i<sch->size; i++) {
        ret = read_schema_elem( fp, sch, i );
        if ( !ret ) return 0;
        }
    return 1;
}

static int read_attribute_details( FILE *fp, attribute *atd )
{
    char * name;
    elem_type type;
    int  size;
    token my_tok;

    my_tok = read_token(fp);
    if ( my_tok != TOK_VAR ) {
        read_error = "expecting attribute name";
        return 0;
        }
    name = mem_strdup(tok_val);
    my_tok = read_token(fp);
    if ( my_tok != TOK_VAR ) {
        read_error = "expecting attribute type";
        return 0;
        }
    if ( tok_is_var("number") ) {
        type = ET_num;
        }
    else if ( tok_is_var("string") ) {
        type = ET_str;
        }
    else {
        read_error = "expecting attribute type";
        return 0;
        }
    my_tok = read_token(fp);
    if ( my_tok != TOK_NUM ) {
        read_error = "expecting attribute size";
        return 0;
        }
    size = tok_num;
    atd->name = name;
    atd->type = type;
    atd->size = size;
    return 1;
}

static int read_schema_elem( FILE *fp, schema *sch, attr_no attr )
{
    attribute atd;
    int ret = read_attribute_details( fp, &atd );
    if ( !ret ) return 0;
    set_attribute_in_schema( sch, attr, atd.name, atd.type, atd.size );
    return 1;
}

/*

int read_table( FILE *fp, table *tab, schema *sch )
{
    int i, ret;
    token my_tok = read_token(fp);
    if ( my_tok != TOK_NUM ) {
        read_error = "expecting table size";
        return 0;
        }
    tab->size = tok_num;
    tab->tups  = (tuple*)mem_malloc( tab->size * sizeof(tuple) );
    for ( i=0; i<tab->size; i++) {
        ret = read_tuple( fp, &tab->tups[i], sch );
        if ( !ret ) return 0;
        }
    return 1;
}

*/

int read_table( FILE *fp, table *tab, schema *sch )
{
    empty_table( tab, sch );
    return read_into_table( fp, tab );
}

int read_into_table( FILE *fp, table *tab )
{
    int i, ret;
    schema *sch = tab->sch;
    token my_tok;

    for ( ; ; ) {
        tuple tup;
        ret = read_tuple( fp, &tup, sch );
        if ( ret == EOF ) break;
        if ( !ret ) return 0;
        add_to_table( tab, tup );
        }
    return 1;
}

static int read_tuple( FILE *fp, tuple *tup, schema *sch )
{
    int i, ret;
    int size = sch->size;
    if ( tok_at_eof(fp) ) return EOF;
    *tup = (tuple_elem*)mem_malloc( size * sizeof(tuple_elem) );
    for (i=0; i<size; i++ ) {
        if ( is_attr_num(i,sch) ) {
            token my_tok = read_token(fp);
            if ( my_tok == TOK_SYM && strcmp(tok_val,"?") ) {
                (*tup)[i].num = null_num;
                }
            else if ( my_tok != TOK_NUM ) {
                read_error = "expecting number";
                return 0;
                }
            else {
                (*tup)[i].num = tok_num;
                }
            }
        else {
            token my_tok = read_token(fp);
            if ( my_tok == TOK_SYM && strcmp(tok_val,"?") ) {
                (*tup)[i].str = null_str;
                }
            else if ( my_tok != TOK_STR && my_tok != TOK_VAR ) {
                read_error = "expecting string";
                return 0;
                }
            else {
                (*tup)[i].str = mem_strdup(tok_val);
                }
            }
        }
    return 1;
}

void print_db( FILE *fp, table *tab )
{
    int i, j;
    currency cur;
    schema *sch = tab->sch;
    for ( i=0; i<sch->size; i++ ) {
        print_str_field(fp, sch->elems[i].size, sch->elems[i].name );
        fputs("  ",fp);
        }
    fputc('\n',fp);
    for ( i=0; i<sch->size; i++ ) {
        print_ch_field(fp, sch->elems[i].size, '=' );
        fputs("  ",fp);
        }
    fputc('\n',fp);

    for ( first_currency( &cur, tab ) ;
                  ! is_last_currency( &cur) ;
                          next_currency( &cur )  ) {
        tuple elem = tuple_at_currency( &cur );
        for ( i=0; i<sch->size; i++ ) {
            elem_type type = sch->elems[i].type;
            int       size = sch->elems[i].size;
            if ( type == ET_num )
                print_num_field(fp, size, elem[i].num );
            else
                print_str_field(fp, size, elem[i].str );
            fputs("  ",fp);
            }
        fputc('\n',fp);
        }
}

void print_table( FILE *fp, table *tab )
{
    int i, j;
    currency cur;
    schema *sch = tab->sch;
    for ( first_currency( &cur, tab ) ;
                  ! is_last_currency( &cur) ;
                          next_currency( &cur )  ) {
        tuple elem = tuple_at_currency( &cur );
        for ( i=0; i<sch->size; i++ ) {
            elem_type type = sch->elems[i].type;
            int       size = sch->elems[i].size;
            if ( type == ET_num )
                print_num_field(fp, size, elem[i].num );
            else
                write_str_field(fp, size, elem[i].str );
            fputs("  ",fp);
            }
        fputc('\n',fp);
        }
}



class_sch *read_class_sch( FILE *fp )
{
    char **name_list;
    int  len, ret;
    attribute atd;
    class_sch *cl;
    ret = read_attribute_details( fp, &atd );
    if ( !ret ) return 0;
    name_list = read_string_list( fp, "{", "}", " ,", &len );
    if ( len < 0 ) return NULL;
    cl = (class_sch *)mem_malloc(sizeof(class_sch));
    cl->det         =  atd;
    cl->nos_classes =  len;
    cl->names       =  name_list;
    return cl;
}

