/**********************************/
/*                                */
/*     dt.h                       */
/*                                */
/*     decision tree              */
/*     definitions                */
/*                                */
/*     Alan Dix                   */
/*                                */
/*     26/11/1993                 */
/*                                */
/**********************************/

#ifndef DT_H
#define DT_H

#include "db.h"

typedef enum { C_leaf, N_const, S_const, S_set, N_pair, S_pair } dec_class;

typedef enum { OP_none, OP_le, OP_lt, OP_eq, OP_in } dec_op;

typedef struct decision_st {
        dec_class class;
        dec_op    op;
        attr_no    attr;
        union dec_info {
            struct {
                  class_no choice;   /*  class chosen at leaf nodes  */
                  int  cf;           /*  certainty factor 0-100%     */
               } c_c;
            struct {
                  int val;           /*  value for constant number tests  */
               } n_c;
            struct {
                  char *val;         /*  value for constant string tests  */
               } s_c;
            struct {
                  int nos_elems;     /*  not used yet  */
                  char **elems;
               } s_s;
            struct {
                  attr_no    att2;   /*  second attribute for paired tests  */
               } pair;
            } inf;
    } decision;

typedef struct dec_l_st {
        struct dec_l_st *next;
        decision dec;
    } *decision_list;

typedef struct t_n {
        decision dec;
        int nos_children;
        struct t_n **children;
    } tree_node;

typedef struct dt_st {
        schema *sch;
        class_sch *cl;
        tree_node *tree;
    } decision_tree;


int nos_choices( decision dec );

int satisfies_cur( decision dec, currency  *cur );
int satisfies( decision dec, tuple tup, schema *sch );
int  classify( table *tab, decision_tree *dt, attr_no attr );
table  *split_classify( table *tab, decision_tree *dt );
int  class_count( table *tab, decision_tree *dt, int *count );
         /*  count[nos_classes]  */
class_no  classify_tuple( tuple tup, decision_tree *dt );

int is_leaf( decision dec );
dec_op lookup_dec_op( char *name );
char *name_of_dec_op( dec_op op );
char *name_of_not_dec_op( dec_op op );

decision_list append_dec_list( decision_list front, decision_list back );
decision_list add_dec_list( decision_list dl, decision dec );
void  free_dec_list( decision_list dl );

decision_list all_possible_decisions( int nos_atts, attr_no *atts, table *tab );

void print_decision_tree( FILE *fp, decision_tree *dt );
void print_decision_tree_as_and_or( FILE *fp, class_no choice, decision_tree *dt );
int read_decision_tree( FILE *fp, decision_tree *dt, schema *sch, class_sch *cl );
void print_dec_list( FILE *fp, decision_list dl, decision_tree *dt );

decision_tree *learn_tree ( class_sch *cl, schema *sch, table *tab, decision_list dl );


#endif /*  DT_H  */

