/**********************************/
/*                                */
/*     gen_io.c                   */
/*                                */
/*     general input/output       */
/*                                */
/*     Alan Dix                   */
/*                                */
/*     26/11/1993                 */
/*                                */
/**********************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "mem.h"

#include "db.h"       /*  for definitions of null_num and null_str  */
#include "gen_io.h"

int token_debug = 0;

char *read_error = NULL;

int  tok_len = 0;
int  tok_num;

static token my_tok = 0;
static token last_tok = 0;

static char tv[TOK_SIZE+1] ="";
static char last_tv[TOK_SIZE+1] ="";
char *tok_val = tv;

static int next_valid( FILE *fp, char *ok_list );
static int do_cvt( char c, char *from, char *to );
static int to_esc( char c );    /*  from character to string form  */
static int from_esc( char c );  /* from escaped character to char code */


int tok_is( char *str )
{
    return strcmp(str,tok_val) == 0;
}

int tok_is_eof()
{
    return my_tok = EOF;
}

int tok_at_eof( FILE *fp )
     /* looks ahead for end of file at next read */
{
    return ( ignore_spaces(fp) == EOF );
}

int tok_is_str()
{
    return my_tok == TOK_STR;
}

int tok_is_num()
{
    return my_tok == TOK_NUM;
}

int tok_is_var( char *str )
{
    return my_tok == TOK_VAR && strcmp(str,tok_val) == 0;
}

int tok_is_sym( char *str )
{
    return my_tok == TOK_SYM && strcmp(str,tok_val) == 0;
}

token unread_token( FILE *fp )
{
    my_tok = last_tok;
    strcpy(tv,last_tv);
    return my_tok;
}

token read_token( FILE *fp )
{
    char c;
    last_tok = my_tok;
    strcpy(last_tv,tv);
    tok_len = 0;
    tok_val[0]='\0';
    ignore_spaces( fp );
    c = fgetc(fp);
    while ( c == '#' ) {
        ignore_line(fp);
        ignore_spaces(fp);
        c = fgetc(fp);
        }
    if ( c==EOF ) {
        my_tok = EOF;
        }
    else if ( isdigit(c) || c=='-' ) {
        int i;
        tok_val[0]=c;
        for( i=1, c=fgetc(fp); i<TOK_SIZE && isdigit(c); i++, c=fgetc(fp) )
                        tok_val[i] = c;
        while ( isdigit(c) ) c = fgetc(fp);
        ungetc(c,fp);
        tok_len = i;
        tok_val[i]='\0';
        tok_num = atoi(tok_val);
        my_tok = TOK_NUM;
        }
    else if ( isalpha(c) || c=='_' ) {
        int i;
        tok_val[0]=c;
        for( i=1, c=fgetc(fp); i<TOK_SIZE && (isalnum(c) || c=='_');
                                                         i++, c=fgetc(fp) )
                        tok_val[i] = c;
        while ( isalnum(c) || c=='_' ) c = fgetc(fp);
        ungetc(c,fp);
        tok_len = i;
        tok_val[i]='\0';
        my_tok = TOK_VAR;
        }
    else if ( c=='\"' ) {
        int i;
        for( i=0, c=fgetc(fp);
                        i<TOK_SIZE && c!=EOF && c!='\"';
                                c=fgetc(fp) ) {
            char esc;
            if ( c != '\\' ) {
                tok_val[i] = c; i++;
                continue;
                }
            c=fgetc(fp);
            if ( c==EOF ) break;
            if ( isdigit(c) ) {
                int val;
                ungetc(c,fp);
                fscanf(fp,"%3o",&val);
                tok_val[i] = val; i++;
                continue;
                }
            if ( c == '\n' ) continue;
            esc = from_esc(c);
            if ( esc )  tok_val[i] = esc;
            else        tok_val[i] = c;
            i++;
            }
        while ( c!=EOF && c!='\"' )  c = fgetc(fp);
        tok_len = i;
        tok_val[i]='\0';
        my_tok = TOK_STR;
        }
    else {
        int i;
        tok_val[0]=c;
        for( i=1, c=fgetc(fp); i<TOK_SIZE && c!=EOF  && !isspace(c) && c!='#'
                                && c!='\"' && c!='-' && !isalnum(c);
                                                         i++, c=fgetc(fp) )
                        tok_val[i] = c;
        while ( c!=EOF  && !isspace(c) && c!='#'
                                && c!='\"' && c!='-' && !isalnum(c) )
                        c = fgetc(fp);
        tok_len = i;
        tok_val[i]='\0';
        my_tok = TOK_SYM;
        }
    return my_tok;
}

static int next_valid( FILE *fp, char *ok_list )
{
    int c;
    char *lp;
    int no_ok = 0;
        
    ignore_spaces(fp);
    c = fgetc(fp);
    for ( lp=ok_list; *lp; lp++ ) {
        if ( *lp == c ) return 1;
        if ( *lp == ' ' ) no_ok = 1;
        }
    ungetc(c,fp);
    return no_ok;
}

char ** read_string_list( FILE *fp, char *start, char *end, char *delim, int *len)
{
    int i, bp, blen = 0;
    char **buff = NULL;
    token tok;
    *len = 0;
    if ( !next_valid(fp,start) ) {
        read_error = "expecting character to start string list";
        return NULL;
        }
    if ( next_valid(fp,end) ) {
        return NULL;
        }
    blen = 10; buff = (char**)mem_malloc( blen * sizeof(char*));
    for ( bp=0, tok = read_token(fp);  ; tok = read_token(fp) ) {
        if ( tok != TOK_STR && tok != TOK_VAR ) {
            read_error = "expecting string";
            break;
            }
        if ( bp >= blen ) {
            blen += 10;
            buff = (char**)mem_realloc( buff, blen * sizeof(char*));
            }
        buff[bp++] = mem_strdup(tok_val);
        if ( next_valid(fp,end) ) {
            buff = (char**)mem_realloc( buff, bp * sizeof(char*));
            *len = bp;
            return buff;
            }
        if ( !next_valid(fp,delim) ) {
            read_error = "expecting delimiter between strings";
            break;
            }
        }
    for ( i=0; i<bp; i++ ) {
        mem_free(buff[bp]);
        }
    free( buff );
    return NULL;
}

void print_indent( FILE *fp, int indent )
{
    int i;
    for (i=0; i<indent; i++) fputc(' ',fp);
}

int ignore_spaces( FILE *fp )
       /* skips spaces and returns lookahead at next character */
       /* returns EOF is the next token would be TOK_EOF       */
{
    char c = fgetc(fp);
    while ( isspace(c) ) c = fgetc(fp);
    ungetc(c,fp);
    return c;
}

void ignore_line( FILE *fp )
{
    char c = fgetc(fp);
    while ( c!=EOF && c!='\n' ) c = fgetc(fp);
}

int lookup( char *n, int size, char**names )
{
    int i;
    for (i=0; i<size; i++) if ( strcmp(n,names[i]) == 0 ) return i;
    return -1;
}

void   print_read_error( FILE *to, FILE *from, char *str)
{
    int c;

    fprintf(to,"error while %s: %s\n",str,read_error);
    fprintf(to,"    when reading %s\n",tok_val);
    fprintf(to,"    context is .... ");
    for ( c = fgetc(from); c!='\n' && c!=EOF; c = fgetc(from) ) fputc(c,to);
    fputc('\n',to);
    fputs("        ",to);
    for ( c = fgetc(from); c!='\n' && c!=EOF; c = fgetc(from) ) fputc(c,to);
    fputc('\n',to);
 }

void print_num_field( FILE *fp, int size, int num )
{
    char fmt[50];
    if ( num == null_num ) {
        print_str_field( fp, size, null_str );
        return;
        }
    sprintf(fmt,"%%%dd",size);
    fprintf(fp,fmt,num);
}

void print_str_field( FILE *fp, int size, char *str )
{
    char fmt[50];
    sprintf(fmt,"%%%ds",size);
    fprintf(fp,fmt,str);
}

int is_var_name( char *str )
{
    char *sp = str;
    if ( *sp == '\0' ) return 1;
    if ( !isalpha(*sp) && !(*sp == '_') ) return 0;
    sp++;
    while ( *sp && ( isalnum(*sp) || *sp=='_' ) ) sp++;
    return *sp=='\0';
}

static char *escape = "\"\'\f\n\t\\";
static char *esc_ch = "\"\'fnt\\";

static int do_cvt( char c, char *from, char *to )
{
    int i;
    for ( i=0;  from[i]; i++ )
        if ( from[i] == c )  return to[i];
    return '\0';
}

static int to_esc( char c )    /*  from character to string form  */
{
    return do_cvt( c, escape, esc_ch );
}

static int from_esc( char c )  /* from escaped character to char code */
{
    return do_cvt( c, esc_ch, escape );
}


void write_str_field( FILE *fp, int size, char *str )
     /*  adds quotes if necessary  */
{
    int ct;
    char *sp;
    char esc;
    if ( is_var_name(str) ) {
        print_str_field( fp, size, str );
        return;
        }
    fputc('\"',fp);
    ct = 1;
    for ( sp=str; *sp; sp++ ) {
        esc = to_esc(*sp);
        if ( esc ) {
            fputc('\\',fp);
            fputc(esc,fp);
            ct += 2;
            }
        else if ( isprint(*sp) ) {
            fputc(*sp,fp);  ct++;
            }
        else {
            fprintf(fp,"\\%03o",*sp);
            ct += 4;
            }
        }
    fputc('\"',fp);
    ct++;
    while ( ct < size ) { 
        fputc(' ',fp);
        ct++;
        }
}

void print_ch_field( FILE *fp, int size, char c )
{
    int i;
    for ( i=0; i<size; i++ ) fputc(c,fp);
}


