/**********************************/
/*                                */
/*     info.c                     */
/*                                */
/*     entropy calculation        */
/*                                */
/*     Alan Dix                   */
/*                                */
/*     2/12/1993                  */
/*                                */
/**********************************/

#include <math.h>
#include "info.h"

#define ln2  0.69314718

static double ent_p( double p )
{
    if ( p < 1.0e-12 ) return 0.0;
    return p * log(p) / ln2;
}

double ent_of_int( int n )
{
    if ( n <= 0 ) return 0.0;
    return n * log(n) / ln2;
}

double ent_of_ints( int size, int *counts )
{
    int i, total;
    double ent;
    for (i=0, ent=0.0; i<size; i++ )
                      ent += ent_of_int( counts[i] );
    return ent;
}

double entropy_from_ints( int size, int *counts )
{
    int i, total;
    double ent;
    for (i=0, total=0; i<size; i++ )  total += counts[i];
    for (i=0, ent=0.0; i<size; i++ )
                      ent += ent_p( ((double)counts[i]) / total );
    return -ent;
}

double entropy( int size, double *prob )
        /* assumes sum of prob[1...size] is 1.0 */
{
    int i;
    double ent;
    for (i=0, ent=0.0; i<size; i++ )  ent += ent_p(prob[i]);
    return -ent;
}



