/**********************************/
/*                                */
/*     learn.c                    */
/*                                */
/*     tree learning              */
/*     code                       */
/*                                */
/*     Alan Dix                   */
/*                                */
/*     3/12/1993                  */
/*                                */
/**********************************/

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "mem.h"

#include "dt.h"
#include "info.h"

#define MAX_DEPTH 3


static double dec_entropy( decision dec, int n_cl, table *tab );
               /*  tab[n_cl]                           */
               /*  all tables should have same schema  */
static tree_node *learn_tree_node ( class_sch *cl, table *tab, decision_list dl, int depth );
static tree_node *do_dec_tree_node ( decision dec, class_sch *cl, table *tab, decision_list dl, int depth );
static tree_node *do_leaf_tree_node ( class_sch *cl, table *tab );
static void split_count( decision dec, table *tab, int *counts );
                      /*  counts[nos_choices(dec)]  */
static void split_table( decision dec, table *tab, table *split );
static int class_ct( int size, table *tabs );
              /*  tabs[size]    */

/* static int tree_depth; */


decision_tree *learn_tree ( class_sch *cl, schema *sch, table *tab, decision_list dl )
               /*  tab[cl->nos_classes]                */
               /*  all tables should have schema sch  */
{
    int nc = cl->nos_classes;
    tree_node *tree;
    decision_tree *dt;
    int i;
#if 0
    decision_list dl;
    int nos_atts;
    attr_no *atts;
#endif

    for ( i=0; i<nc; i++ ) assert( equal_schema(tab[i].sch,sch) );

#if 0        
    nos_atts = sch->size;
    atts = (attr_no *)mem_malloc( nos_atts * sizeof(attr_no) );
    for (i=0; i<nos_atts; i++ ) atts[i] = i;
    dl = all_possible_decisions( nos_atts, atts, tab );
#endif

/*    tree_depth = 0;  */

    tree = learn_tree_node ( cl, tab, dl, 1 );
    if ( tree == NULL ) return NULL;

    dt = (decision_tree *)mem_malloc(sizeof(decision_tree));
    dt->cl = cl;
    dt->sch = tab->sch;
    dt->tree = tree;
    return dt;
}

#define  MAX_NOS_CLASSES  10
#define  MAX_DEC_CHOICES  10

static double dec_entropy( decision dec, int n_cl, table *tab )
               /*  tab[n_cl]                           */
               /*  all tables should have same schema  */
{
static int counts[MAX_NOS_CLASSES][MAX_DEC_CHOICES];
static int cl_cts[MAX_NOS_CLASSES];
static int tot_cts[MAX_DEC_CHOICES];
    int tot, i, j;
    double n_ent = 0.0,
           cl_ent = 0.0,
           ch_ent = 0.0,
           all_ent = 0.0;
    int    d_ch = nos_choices(dec);
    assert( d_ch < MAX_DEC_CHOICES );
    assert( n_cl < MAX_NOS_CLASSES );

    tot = 0;
    for( j=0; j<d_ch; j++ ) tot_cts[j] = 0;
    for ( i=0; i<n_cl; i++ ) {
        tot += tab[i].size;
        cl_cts[i] = tab[i].size;
        split_count( dec, &tab[i], counts[i] );
        for( j=0; j<d_ch; j++ ) tot_cts[j] += counts[i][j];
        all_ent += ent_of_ints( d_ch, counts[i] );
        }
    n_ent  = ent_of_int( tot );
    cl_ent = ent_of_ints( n_cl, cl_cts );
    ch_ent = ent_of_ints( d_ch, tot_cts );
/*    return  all_ent - ch_ent;  */
         /*  alternate measure  */
    return  (all_ent - cl_ent - ch_ent + n_ent) / tot;
}

#define CRITICAL_ENT 0.00001

static tree_node *learn_tree_node ( class_sch *cl, table *tab, decision_list dl, int depth )
               /*  tab[cl->nos_classes]                */
               /*  all tables should have same schema  */
{
    decision_list this;
    decision dec;
    double best_ent = 0;
    int n_cl = cl->nos_classes;
    int ct = class_ct( n_cl, tab );

/*    tree_depth++;                                               */
/*    if ( tree_depth > 2 ) return do_leaf_tree_node( cl, tab );  */
    if ( depth > MAX_DEPTH || ct <= 1 ) return do_leaf_tree_node( cl, tab );

    for ( this=dl; this; this=this->next ) {
        double ent = dec_entropy ( this->dec, n_cl, tab );
        if ( ent > best_ent ) {
            best_ent = ent;
            dec = this->dec;
            }
        }
    if ( best_ent > CRITICAL_ENT )
         return do_dec_tree_node( dec, cl, tab, dl, depth+1 );
    else return do_leaf_tree_node( cl, tab );
}

static tree_node *do_dec_tree_node ( decision dec, class_sch *cl, table *tab, decision_list dl, int depth )
               /*  tab[cl->nos_classes]                */
               /*  all tables should have same schema  */
{
    int i;
    int d_ch = nos_choices(dec);
    tree_node *tn = (tree_node*)mem_malloc(sizeof(tree_node));

    int j;
    int n_cl = cl->nos_classes;
    table **tabs;
    table **tabs_trans;
 
    tn->dec = dec;
    tn->nos_children = d_ch;
    tn->children = (tree_node**)mem_malloc( d_ch * sizeof(tree_node*) );

#if 0
    for ( i=0; i<d_ch; i++ ) {
        tn->children[i] = do_leaf_tree_node( cl, tab );
        }
    return tn;
#endif

    tabs       =  (table**) mem_malloc( n_cl * sizeof(table*) );
    tabs_trans =  (table**) mem_malloc( d_ch * sizeof(table*) );
    for ( i=0; i<n_cl; i++ )
        tabs[i]        =  (table*) mem_malloc( d_ch * sizeof(table) );
    for ( j=0; j<d_ch; j++ )
        tabs_trans[j]  =  (table*) mem_malloc( n_cl * sizeof(table) );

    for ( i=0; i<n_cl; i++ ) {
        split_table( dec, &tab[i], tabs[i] );
        }
    for ( i=0; i<n_cl; i++ ) {
        for( j=0; j<d_ch; j++ ) tabs_trans[j][i] = tabs[i][j];
        }
    for ( j=0; j<d_ch; j++ ) {
/*        tn->children[j] = do_leaf_tree_node( cl, tabs_trans[j] ); */
        tn->children[j] = learn_tree_node ( cl, tabs_trans[j], dl, depth );
        }

    for ( i=0; i<n_cl; i++ ) mem_free(tabs[i]);
    for ( j=0; j<d_ch; j++ )  mem_free(tabs_trans[j]);
    mem_free(tabs); mem_free(tabs_trans);

    return tn;
 
}

static tree_node *do_leaf_tree_node ( class_sch *cl, table *tab )
               /*  tab[cl->nos_classes]                */
               /*  all tables should have same schema  */
{
    int n_cl = cl->nos_classes;
    int i, best_cl, best_ct, tot;
    double certainty_factor;
    tree_node *tn = (tree_node*)mem_malloc(sizeof(tree_node));
    tn->dec.class = C_leaf;
    tn->dec.op    = OP_none;
    tn->nos_children = 0;
    tn->children = NULL;
    if ( n_cl <= 0 ) {
        tn->dec.inf.c_c.choice = 0;
        tn->dec.inf.c_c.cf     = 0;
        return tn;
        }
    best_cl = 0;
    best_ct = tab[0].size;
    tot     = best_ct;
    for( i=1; i<n_cl; i++ ) {
        tot += tab[i].size;
        if ( tab[i].size > best_ct ) {
            best_ct = tab[i].size;
            best_cl = i;
            }
        }
    if ( tot == 0 ) {
        tn->dec.inf.c_c.choice = 0;
        tn->dec.inf.c_c.cf     = 0;
        }
    else {
        tn->dec.inf.c_c.choice = best_cl;
        tn->dec.inf.c_c.cf     = (100*best_ct)/tot;
        }
    return tn;
}


static void split_count( decision dec, table *tab, int *counts )
                      /*  counts[nos_choices(dec)]  */
{
    int i, j;
    currency cur;
    schema *sch = tab->sch;
    int n_c = nos_choices(dec);
    for ( i=0; i<n_c; i++ ) counts[i] = 0;

    for ( first_currency( &cur, tab ) ;
                  ! is_last_currency( &cur) ;
                          next_currency( &cur )  ) {
        tuple elem = tuple_at_currency( &cur );
        int   ch = satisfies_cur( dec, &cur );
        counts[ch]++;
        }
}

static void split_table( decision dec, table *tab, table *split )
                      /*  splits[nos_choices(dec)]  */
{
    int i;
    currency cur;
    schema *sch = tab->sch;
    
    int n_c = nos_choices(dec);
    for ( i=0; i<n_c; i++ ) {
        schema *cpy = copy_schema(sch);
        empty_table(&split[i],cpy);
        }

    for ( first_currency( &cur, tab ) ;
                  ! is_last_currency( &cur) ;
                          next_currency( &cur )  ) {
        int   ch = satisfies_cur( dec, &cur );
        add_cur_to_table( &split[ch], &cur );
        }
}

static int class_ct( int size, table *tabs )
              /*  tabs[size]    */
{
    int i;
    int ct = 0;
    for ( i=0; i<size; i++ )
        if ( tabs[i].size > 0 ) ct++;
    return ct;
}

