/**********************************/
/*                                */
/*     main.c                     */
/*                                */
/*     inductive learning         */
/*     test program               */
/*                                */
/*     Alan Dix                   */
/*                                */
/*     26/11/1993                 */
/*                                */
/**********************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <assert.h>

#include "mem.h"

#include "db.h"
#include "dt.h"

#define PROGNAME_SIZE 30
static char progname[PROGNAME_SIZE+1] = "";

static void set_progname(char *path);
static int is_progname(char *name);

char *c_names[2] = { "yes", "no" };

class_sch my_class = { {"class", ET_str, 5}, 2, c_names };

schema_elem elems[4] = {
    { "name",     ET_str,  15 },
    { "balance",  ET_num,  15 },
    { "contact",  ET_str,  15 },
    { "limit",    ET_num,  15 }
  };

schema my_schema = { 4, elems };

decision_tree my_tree;

attribute  ext_attrib = { "class",  ET_str,  5 };
tuple_elem ext_init;


FILE *my_fopen(char *name,char *mode)
{
    FILE *fp = fopen(name,mode);
    if ( fp == NULL ) {
        fprintf(stderr,"couldn't open %s in mode %s\n",name,mode);
        exit(1);
        }
    return fp;
}

static int test( int argc, char **argv );
static int learn( int argc, char **argv );
static int asandor( int argc, char **argv );
static int class( int argc, char **argv );
static int split( int argc, char **argv );

static int parse_args(char *buff, char ***argv);


static FILE *read_open(char *name)
{
    FILE *fp = fopen(name,"r");
    if ( fp ) return fp;
    fprintf(stderr,"couldn't open %s for reading\n",name);
    exit(1);
}

static FILE *write_open(char *name)
{
    FILE *fp = fopen(name,"w");
    if ( fp ) return fp;
    fprintf(stderr,"couldn't open %s for writing\n",name);
    exit(1);
}

typedef struct {
    char *name;
    int (*f)( int argc, char **argv );
    char *usage;
    } pt_st;

static pt_st prog_tab[6] = {
   { "test",    test,     "various test runs"      },
   { "learn",   learn,    "attempt_nos schema class tree(OUT) tab1 ..."  },
   { "asandor", asandor,  "schema class tree(IN) class_to_choose  -  and/or expression to stdout" },
   { "class",   class,    "schema class tree(IN) table  -  classification to stdout" },
   { "split",   split,    "schema class tree(IN) table tab1(OUT) ..." },
   { NULL,      NULL ,    NULL    }
   };

static char *usage_string = "none";

static void usage()
{
    fprintf(stderr,"usage: %s %s\n",progname,usage_string);
    exit(1);
}

static void valid_names()
{
}

char *default_argv[2] = {"main",NULL};

main(argc,argv)
    int argc;
    char **argv;
{
    int ac = argc;
    char **av = argv;
    pt_st *pt;
#if 0
    int i;
    printf("main(%d,%08x)\n",argc,argv);
    for (i=0; i<argc; i++ )
    	printf("    argv[%d] = \"%s\"\n",i,argv[i]);
    exit(1);
#endif
    if ( ac <= 0 ) { ac=1; av=default_argv; }
    set_progname(av[0]);
    if ( ac == 1 || is_progname("main") ) {
        ac--; av++;
        if ( ac <= 0 ) {
            char buff[1000];
            gets(buff);    /* beware very long lines!!!  */
            ac=parse_args(buff,&av);
            }
        if ( ac <=0 ) usage();
        set_progname(av[0]);
        }
    for ( pt = prog_tab; pt->name != NULL; pt++ )
        if ( is_progname(pt->name) ) {
            usage_string = pt->usage;
            return (pt->f)(ac,av);
            }
    fprintf(stderr,"unrecognised program name: %s\n",progname);
    fprintf(stderr,"valid names:\n");
    for ( pt = prog_tab; pt->name != NULL; pt++ ) {
        fprintf(stderr,"     %10s",pt->name);
        if ( pt->usage && *(pt->usage) )
                fprintf(stderr,"   -   %s\n",pt->usage);
        else    fputc('\n',stderr);
        }
    return 1;
}

#define MAX_ARGS 20
static char *my_argv[MAX_ARGS+1];

static int parse_args(char *buff, char ***argv)
{
    int argc;
    char *bp;
    enum { space, nospace, none } mode = none;
    for ( argc=0, bp=buff; *bp; bp++ ) {
        if ( mode!=space && isspace(*bp) ) {
            *bp = '\0';
            mode = space;
            }
        else if ( mode!=nospace && !isspace(*bp) ) {
            if ( argc < MAX_ARGS ) my_argv[argc++] = bp;
            mode = nospace;
            }
        }
    *argv = my_argv;
    my_argv[argc] = NULL;
    return argc;
}

static void learn_usage()
{
    fprintf(stderr,"usage: learn attempt_nos schema classes tree(OUT) tab1 ...\n");
    exit(1);
}

static int learn( int argc, char **argv )
{
    char *schema_name, *class_name, *tree_name, **tab_names;
    FILE *sch_fp, *cl_fp, *tree_fp;
    schema sch;
    class_sch *cl;
    int ret, i, nos, nos_tabs;
    table *tabs;
    decision_tree *new_tree;
    decision_list dl;
    int nos_atts;
    attr_no *atts;
    table my_table;
    int attempt_nos;


    if ( argc < 6 )  learn_usage();
    if ( ! isdigit(*argv[1]) ) {
        fprintf( stderr, "expected number for attempt nos, not %s\n",
                          argv[1] );
        learn_usage();
        exit(1);
        }
    attempt_nos = atoi(argv[1]);
    schema_name = argv[2];
    class_name  = argv[3];
    tree_name   = argv[4];
    tab_names   = &argv[5];
    nos_tabs = argc-5;
    sch_fp = read_open(schema_name);
    cl_fp  = read_open(class_name);
    ret  = read_schema( sch_fp, &sch );
    if ( ! ret ) {
        print_read_error( stderr, sch_fp, "reading schema" );
        exit(1);
        }
    cl = read_class_sch( cl_fp );
    if ( ! cl ) {
        print_read_error( stderr, cl_fp, "reading class schema" );
        exit(1);
        }
    nos = cl->nos_classes;
    if ( nos_tabs != nos ) {
        fprintf(stderr,"%s: wrong number of tables, need %d\n",progname,nos);
        learn_usage();
        }
    tabs = (table *)mem_malloc( nos * sizeof(table) );
    empty_table( &my_table, &sch );
    for ( i=0; i<nos; i++ ) {
        FILE *fp = read_open(tab_names[i]);
        ret = read_table( fp, &tabs[i], &sch );
        if ( ! ret ) {
        print_read_error( stderr, fp, "reading table" );
        exit(1);
        }
        append_table( &my_table, &tabs[i] );
        }

    nos_atts = my_schema.size;
    atts = (attr_no *)mem_malloc( nos_atts * sizeof(attr_no) );
    for ( i=0; i<nos_atts; i++ )   atts[i] = i; 
    dl = all_possible_decisions( nos_atts, atts, &my_table );

    new_tree = learn_tree ( cl, &sch, tabs, dl );
    if ( strcmp(tree_name,"-") == 0 )
          tree_fp = stdout;
    else  tree_fp = write_open(tree_name);
    print_decision_tree( tree_fp, new_tree );
    fputc('\n',tree_fp);
}

static int not_implemented( int argc, char **argv )
{
    fprintf(stderr,"%s not yet implemented\n",progname);
    exit(1);
}

static int asandor( int argc, char **argv )
{
    char *schema_name, *class_name, *tree_name, *class_choice;
    FILE *sch_fp, *cl_fp, *tree_fp;
    schema sch;
    class_sch *cl;
    int ret, i, nos;
    table in_table;
    decision_tree my_tree;
    class_no class_idx;

    if ( argc != 5 )  usage();
    schema_name = argv[1];
    class_name  = argv[2];
    tree_name   = argv[3];
    class_choice  = argv[4];
    sch_fp   = read_open(schema_name);
    cl_fp    = read_open(class_name);
    tree_fp  = read_open(tree_name);
    ret  = read_schema( sch_fp, &sch );
    if ( ! ret ) {
        print_read_error( stderr, sch_fp, "reading schema" );
        exit(1);
        }
    cl = read_class_sch( cl_fp );
    if ( ! cl ) {
        print_read_error( stderr, cl_fp, "reading class schema" );
        exit(1);
        }

	class_idx = get_class_no( class_choice, cl );
    if ( class_idx < 0 ) {
        fprintf( stderr, "bad class name %s", class_choice );
        exit(1);
        }

    ret = read_decision_tree( tree_fp, &my_tree, &sch, cl );
    if ( ! ret ) {
        print_read_error( stderr, tree_fp, "reading decision tree" );
        exit(1);
        }

    nos = cl->nos_classes;
    
    print_decision_tree_as_and_or( stdout, class_idx, &my_tree );
    printf("\n");

    /*
    ext_init.str = NULL;
    ext_tab = add_attrib_to_table( &in_table, cl->det, ext_init );
    ext_attr_no = get_attr_no( cl->det.name, ext_tab->sch );
    ret = classify( ext_tab, &my_tree, ext_attr_no );
    if ( ! ret ) exit(1);
    print_db( stdout, ext_tab );
    */

    return 0;
}

static int class( int argc, char **argv )
{
    char *schema_name, *class_name, *tree_name, *table_name, **tab_names;
    FILE *sch_fp, *cl_fp, *tree_fp,  *tb_fp;
    schema sch;
    class_sch *cl;
    int ret, i, nos;
    table in_table;
    decision_tree my_tree;
    table  *ext_tab;
    attr_no ext_attr_no;

    if ( argc != 5 )  usage();
    schema_name = argv[1];
    class_name  = argv[2];
    tree_name   = argv[3];
    table_name  = argv[4];
    sch_fp   = read_open(schema_name);
    cl_fp    = read_open(class_name);
    tree_fp  = read_open(tree_name);
    tb_fp    = read_open(table_name);
    ret  = read_schema( sch_fp, &sch );
    if ( ! ret ) {
        print_read_error( stderr, sch_fp, "reading schema" );
        exit(1);
        }
    cl = read_class_sch( cl_fp );
    if ( ! cl ) {
        print_read_error( stderr, cl_fp, "reading class schema" );
        exit(1);
        }

    ret = read_table( tb_fp, &in_table, &sch );
    if ( ! ret ) {
        print_read_error( stderr, tb_fp, "reading table" );
        exit(1);
        }
    ret = read_decision_tree( tree_fp, &my_tree, &sch, cl );
    if ( ! ret ) {
        print_read_error( stderr, tree_fp, "reading decision tree" );
        exit(1);
        }

    nos = cl->nos_classes;

    ext_init.str = NULL;
    ext_tab = add_attrib_to_table( &in_table, cl->det, ext_init );
    ext_attr_no = get_attr_no( cl->det.name, ext_tab->sch );
    ret = classify( ext_tab, &my_tree, ext_attr_no );
    if ( ! ret ) exit(1);
    print_db( stdout, ext_tab );

    return 0;
}

static int split( int argc, char **argv )
{
    char *schema_name, *class_name, *tree_name, *table_name, **tab_names;
    FILE *sch_fp, *cl_fp, *tree_fp,  *tb_fp;
    schema sch;
    class_sch *cl;
    int ret, i, nos, nos_tabs;
    table in_table;
    table *splits;
    decision_tree my_tree;

    if ( argc < 6 )  usage();
    schema_name = argv[1];
    class_name  = argv[2];
    tree_name   = argv[3];
    table_name  = argv[4];
    tab_names   = &argv[5];
    nos_tabs = argc-5;
    sch_fp   = read_open(schema_name);
    cl_fp    = read_open(class_name);
    tree_fp  = read_open(tree_name);
    tb_fp    = read_open(table_name);
    ret  = read_schema( sch_fp, &sch );
    if ( ! ret ) {
        print_read_error( stderr, sch_fp, "reading schema" );
        exit(1);
        }
    cl = read_class_sch( cl_fp );
    if ( ! cl ) {
        print_read_error( stderr, cl_fp, "reading class schema" );
        exit(1);
        }

    ret = read_table( tb_fp, &in_table, &sch );
    if ( ! ret ) {
        print_read_error( stderr, tb_fp, "reading table" );
        exit(1);
        }
    ret = read_decision_tree( tree_fp, &my_tree, &sch, cl );
    if ( ! ret ) {
        print_read_error( stderr, tree_fp, "reading decision tree" );
        exit(1);
        }

    nos = cl->nos_classes;
    if ( nos_tabs != nos ) {
        fprintf(stderr,"%s: wrong number of tables, need %d\n",progname,nos);
        usage();
        }
    splits = split_classify( &in_table, &my_tree );
    for ( i=0; i<nos_tabs; i++ ) {
        FILE *fp = write_open(tab_names[i]);
        print_table( fp, &splits[i] );
        }
    return 0;
}

static int test( int argc, char **argv )
{
    FILE *dt_fp, *sch_fp, *db_fp;
    int ret;
    table  *ext_tab;
    attr_no ext_attr_no;
    decision_list dl;
    int nos_atts, i;
    attr_no *atts;
    table *splits;
    decision_tree *new_tree;
    table my_table;

    if ( argc != 4 ) {
        fprintf(stderr,"usage: main schema dbfile tree\n");
        exit(1);
        }

    mem_init();

    sch_fp = my_fopen(argv[1],"r");
    db_fp  = my_fopen(argv[2],"r");
    dt_fp  = my_fopen(argv[3],"r");
    ret  = read_schema( sch_fp, &my_schema );
    ret &= read_table( db_fp, &my_table, &my_schema );
    if ( ! ret ) {
        print_read_error( stderr, db_fp, "reading table" );
        exit(1);
        }
    print_db( stdout, &my_table );
    printf("\n");
/*    exit(0);  */
    ret &= read_decision_tree( dt_fp, &my_tree, &my_schema, &my_class );
    if ( ! ret ) {
        print_read_error( stderr, dt_fp, "reading decision tree" );
        exit(1);
        }
    print_decision_tree( stdout, &my_tree );
    printf("\n");
    ext_init.str = NULL;
    ext_tab = add_attrib_to_table( &my_table, ext_attrib, ext_init );
    ext_attr_no = get_attr_no( "class", ext_tab->sch );
    ret = classify( ext_tab, &my_tree, ext_attr_no );
    if ( ! ret ) exit(1);
    print_db( stdout, ext_tab );
    nos_atts = my_schema.size;
    atts = (attr_no *)mem_malloc( nos_atts * sizeof(attr_no) );
    for ( i=0; i<nos_atts; i++ )   atts[i] = i; 
    dl = all_possible_decisions( nos_atts, atts, &my_table );
    print_dec_list( stdout, dl, &my_tree );
    splits = split_classify( &my_table, &my_tree );
    for ( i=0; i<2; i++ ) {
        printf("\nCLASS %3s\n=========\n",c_names[i]);
        print_db( stdout, &splits[i] );
        }
    new_tree = learn_tree ( &my_class, &my_schema, splits, dl );
    printf("\n");
    print_decision_tree( stdout, new_tree );
    printf("\n");

    /* mem_term(); */   /*  always gives error because  */
                        /*  I don't free everything!    */
    exit(0);
}

static int is_progname(char *name)
{
    return strcmpl(name,progname) == 0;
}

static void set_progname(char *path)
{
    int i, j, last_slash, dot;
    if ( path == NULL ) {
        progname[0] = '\0';
        return;
        }
    for ( i=0, last_slash=0, dot=0;
                  path[i] != '\0';  i++  ) {
        if ( path[i] == '\\' || path[i] == '/' ) last_slash = i+1;
        if ( path[i] == '.' ) dot = i+1;
        }
    for ( i=0, j=last_slash;
                i<PROGNAME_SIZE && path[j] && path[j] != '.';   i++, j++ )
            progname[i] = tolower(path[j]);
    progname[i] = '\0';
}
